#ifndef MSP_GL_MESH_H_
#define MSP_GL_MESH_H_

#include <msp/datafile/objectloader.h>
#include "batch.h"
#include "resource.h"
#include "vertexarray.h"
#include "windingtest.h"

namespace Msp {
namespace GL {

class Buffer;
class Renderer;
class VertexSetup;

/**
Raw mesh data, consisting of a VertexArray and one or more Batches.  Though a
Mesh can draw itself, it's usually used as part of Renderables rather than on
its own.
*/
class Mesh: public Bindable<Mesh>, public Resource
{
	friend class MeshBuilder;

public:
	class Loader: public DataFile::ObjectLoader<Mesh>
	{
	public:
		Loader(Mesh &);
	private:
		void vertices(const std::vector<VertexComponent> &);
		void batch(PrimitiveType);
		void winding(FaceWinding);
	};

private:
	class AsyncLoader: public Resource::AsyncLoader
	{
	private:
		Mesh &mesh;
		IO::Seekable &io;
		Bufferable::AsyncUpdater *vertex_updater;
		Bufferable::AsyncUpdater *index_updater;
		unsigned phase;

	public:
		AsyncLoader(Mesh &, IO::Seekable &);
		~AsyncLoader();

		virtual bool needs_sync() const;
		virtual bool process();
	};

	VertexArray vertices;
	std::vector<Batch> batches;
	Buffer *vbuf;
	Buffer *ibuf;
	VertexSetup *vtx_setup;
	bool defer_buffers;
	mutable bool dirty;
	bool disallow_rendering;
	const WindingTest *winding;

public:
	Mesh(ResourceManager * = 0);
	Mesh(const VertexFormat &, ResourceManager * = 0);
private:
	void init(ResourceManager *);
public:
	~Mesh();

	void clear();
	void use_buffers(bool);
private:
	void create_buffers();

public:
	const VertexArray &get_vertices() const { return vertices; }
	const VertexSetup *get_vertex_setup() const { return vtx_setup; }
	const Buffer *get_index_buffer() const { return ibuf; }
	unsigned get_n_vertices() const;
	float *modify_vertex(unsigned);

	void add_batch(const Batch &b);
	const std::vector<Batch> &get_batches() const { return batches; }

	void set_winding(const WindingTest *);

	void draw() const;
	void draw(Renderer &) const;
	void draw_instanced(Renderer &, const VertexSetup &, unsigned) const;

	/** Binds the mesh for rendering.  The vertex array is applied using generic
	attributes only.  Uses vertex array object if possible. */
	void bind() const;

	static void unbind();

	virtual int get_load_priority() const { return 1; }
	virtual Resource::AsyncLoader *load(IO::Seekable &, const Resources * = 0);
	virtual UInt64 get_data_size() const;
	virtual void unload();
};

} // namespace GL
} // namespace Msp

#endif
